;; Generated package description from inf-clojure.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "inf-clojure" "3.4.0.0.20260507.60144" "Basic interaction with a Clojure REPL" '((emacs "28.1") (clojure-mode "5.11")) :commit "8fff36b7cb2919ca2b64f2cfb0caf6406511d3af" :maintainer '("Bozhidar Batsov" . "bozhidar@batsov.dev") :keywords '("processes" "comint" "clojure") :url "https://github.com/clojure-emacs/inf-clojure")
